from dataclasses import fields, make_dataclass, is_dataclass
from typing import List, Tuple
from typing import Optional, get_type_hints, Dict, Type, Any

_partial_cache: Dict[Type[Any], Type[Any]] = {}


def make_dataclass_with_optional_paths(cls: Type[Any]) -> Type[Any]:
    assert is_dataclass(cls), f"{cls} must be a dataclass"

    new_fields: List[Tuple[str, Any]] = []
    hints = get_type_hints(cls)

    for f in fields(cls):
        field_type = hints[f.name]

        # If it's a dataclass, add both the field and its '_path_to_yaml_file' sibling
        if is_dataclass(field_type):
            new_fields.append((f.name, Optional[field_type]))
            new_fields.append((f"{f.name}_path_to_yaml_file", Optional[str]))
        else:
            new_fields.append((f.name, field_type))

    new_cls_name = cls.__name__ + "_with_potential_path"
    return make_dataclass(new_cls_name, new_fields)


def make_partial_dataclass(cls: Type[Any]) -> Type[Any]:
    if cls in _partial_cache:
        return _partial_cache[cls]

    if not is_dataclass(cls):
        raise ValueError(f"{cls} is not a dataclass")

    type_hints = get_type_hints(cls)
    partial_fields = []

    for field in fields(cls):
        field_type = type_hints[field.name]

        # If it's a nested dataclass, recurse
        if is_dataclass(field_type):
            assert isinstance(field_type, type)
            partial_type = make_partial_dataclass(field_type)
        else:
            partial_type = field_type

        # Make the field optional
        partial_fields.append(
            (
                field.name,
                Optional[partial_type],
                field.default if field.default != field.default_factory else None,
            )
        )

    # Create the new Partial dataclass
    partial_cls_name = f"Partial{cls.__name__}"

    annotations = {name: typ for name, typ, *_ in partial_fields}
    partial_cls = make_dataclass(
        partial_cls_name, partial_fields, namespace={"__annotations__": annotations}
    )

    _partial_cache[cls] = partial_cls
    return partial_cls


def make_partial_dataclass_with_optional_paths(cls: Type[Any]) -> Type[Any]:
    optional_paths_class = make_dataclass_with_optional_paths(cls=cls)
    partial_with_optional_paths_class = make_partial_dataclass(cls=optional_paths_class)
    return partial_with_optional_paths_class



"""Imports the main class from the parlsey module."""

from parsley_coco.parser import Parsley
from parsley_coco.factory import create_parsley
from parsley_coco.alternative_dataclasses import (
    make_partial_dataclass_with_optional_paths,
)


___all__ = ["Parlsey", "create_parsley", "make_partial_dataclass_with_optional_paths"]



from dataclasses import fields, is_dataclass, asdict
from enum import Enum
from typing import Any, Type

import dacite
import yaml
from dacite import from_dict

from parsley_coco.alternative_dataclasses import make_dataclass_with_optional_paths
from parsley_coco.utils import IsDataclass


def resolve_yaml_to_base[T_Dataclass: IsDataclass](
    yaml_path: str, base_cls: Type[T_Dataclass]
) -> T_Dataclass:
    extended_cls = make_dataclass_with_optional_paths(base_cls)

    try:
        with open(yaml_path, "r") as f:
            yaml_data = yaml.safe_load(f)
    except IOError:
        raise Exception("Could not read file:", yaml_path)

    extended_obj = from_dict(
        data_class=extended_cls, data=yaml_data, config=dacite.Config(cast=[Enum])
    )

    return resolve_extended_object(extended_obj, base_cls)


def resolve_extended_object_to_dict[T_Dataclass: IsDataclass](
    extended_obj: IsDataclass, base_cls: Type[T_Dataclass], raise_error_with_nones: bool = True
) -> dict[str, Any]:
    resolved_data = {}

    assert is_dataclass(extended_obj)
    for field in fields(base_cls):
        base_field_type = field.type
        val = getattr(extended_obj, field.name, None)

        if is_dataclass(base_field_type):
            path_val = getattr(extended_obj, f"{field.name}_path_to_yaml_file", None)

            if path_val and not val:
                assert isinstance(base_field_type, type)
                resolved_val = asdict(resolve_yaml_to_base(path_val, base_field_type))
            elif val and not path_val:
                assert isinstance(base_field_type, type)
                resolved_val = resolve_extended_object_to_dict(val, base_field_type)
            else:
                if raise_error_with_nones:
                    string = f"Exactly one of the fields {field.name} or {field.name}_path_to_yaml_file must be given, "
                    if not val and not path_val:
                        string += "not none of them"
                    elif val is not None and path_val is not None:
                        string += "not both"
                    raise ValueError(string)
                else:
                    resolved_val = None

            resolved_data[field.name] = resolved_val

        else:
            resolved_data[field.name] = val

    return resolved_data


def resolve_extended_object[T_Dataclass: IsDataclass](
    extended_obj: Any, base_cls: Type[T_Dataclass]
) -> T_Dataclass:
    resolved_data: dict[str, Any] = resolve_extended_object_to_dict(
        extended_obj=extended_obj, base_cls=base_cls
    )

    return from_dict(
        data_class=base_cls, data=resolved_data, config=dacite.Config(cast=[Enum])
    )
